/* 
 * Copyright (C) 2004-2005 Jonathan Bindel
 * Copyright (C) 2006-2007 Eskil Bylund
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

using System;
using System.Collections.Generic;
using System.Xml.Serialization;

namespace DCSharp.Backend.Objects
{
	public enum Priority
	{
		High,
		Normal,
		Low
	}
	
	public class DownloadEventArgs : EventArgs
	{
		public DownloadEventArgs(DownloadFileInfo download)
		{
			this.download = download;
		}
		
		public DownloadFileInfo Download
		{
			get { return download; }
		}
		private DownloadFileInfo download;
	}
	
	public class DownloadFileInfo : TransferFileInfo
	{
		public event EventHandler<SourceEventArgs> SourceAdded;
		public event EventHandler<SourceEventArgs> SourceRemoved;
		
		#region Constructors
		
		public DownloadFileInfo(string cid) : this(cid, 0, null, null)
		{
			IsFileList = true;
		}
		
		public DownloadFileInfo(string name, long size, string tth, string target) :
			base(name, size, tth)
		{
			this.target = target;
			tempTarget = target;
			
			sources = new List<SourceInfo>();
			priority = Priority.Normal;
		}
		
		public DownloadFileInfo() : base()
		{
			sources = new List<SourceInfo>();
			priority = Priority.Normal;
		}

		#endregion

		#region Properties
		
		[XmlAttribute]
		public string Target
		{
			get { return target; }
			set { target = value; }
		}
		private string target;
		
		[XmlAttribute]
		public string TempTarget
		{
			get { return tempTarget; }
			set { tempTarget = value; }
		}
		private string tempTarget;
		
		[XmlElement("Source")]
		public SourceInfo[] SourceInfos
		{
			get { return sources.ToArray(); }
			set
			{
				sources.Clear();
				if (value != null)
				{
					sources.AddRange(value);
				}
			}
		}
		private List<SourceInfo> sources;
		
		[XmlAttribute]
		public Priority Priority
		{
			get { return priority; }
			set { priority = value; }
		}
		private Priority priority;
		
		#endregion

		#region Methods

		public void Add(SourceInfo source)
		{
			if (source == null)
			{
				throw new ArgumentNullException("source");
			}
			sources.Add(source);
			OnSourceAdded(source);
		}

		public void Remove(SourceInfo source)
		{
			if (source == null)
			{
				throw new ArgumentNullException("source");
			}
			if (sources.Contains(source))
			{
				if (Active && source.Active)
				{
					UserConnection.Disconnect();
				}
				sources.Remove(source);
				OnSourceRemoved(source);
			}
		}
		
		public bool Contains(SourceInfo source)
		{
			return sources.Contains(source);
		}

		public bool HasSource(SourceInfo source)
		{
			return HasSource(source.User, source.Path);
		}

		public bool HasSource(User user)
		{
			return HasSource(user, null);
		}
		
		public bool HasSource(User user, string path)
		{
			SourceInfo source = sources.Find(delegate(SourceInfo s)
			{
				return s.Available && s.User == user &&
					(path == null || s.Path == path);
			});
			return source != null;
		}

		protected virtual void OnSourceAdded(SourceInfo source)
		{
			if (SourceAdded != null)
			{
				SourceAdded(this, new SourceEventArgs(source));
			}
		}

		protected virtual void OnSourceRemoved(SourceInfo source)
		{
			if (SourceRemoved != null)
			{
				SourceRemoved(this, new SourceEventArgs(source));
			}
		}

		#endregion
	}
}
